<?php
namespace Util\Events\Task;

use Doctrine\Common\EventSubscriber;
use Doctrine\ORM\Event\LifecycleEventArgs;
use Doctrine\ORM\Event\PostFlushEventArgs;
use Doctrine\ORM\Event\PreUpdateEventArgs;
use Doctrine\ORM\Events;
use Zeedhi\DevOps\Model\Projectmanagement\Entities\Task as TaskEntity;
use Zeedhi\DevOps\Model\Projectmanagement\Entities\Allocation as AllocationEntity;
use Zeedhi\Framework\DependencyInjection\InstanceManager;

class IssueSubscriber implements EventSubscriber
{
    /***********************************************************************
     * Entity fields names
     ***********************************************************************/
    const DESCRIPTION_FIELD = "description";
    const TASK_STATUS_FIELD = "taskStatus";
    const REPORTER_FIELD    = "resource";

    /***********************************************************************
     * Array containing issue changes.
     ***********************************************************************/
    private $issueChangeSet;

    /***********************************************************************
     * Controls flush events avoiding an infinite loop.
     ***********************************************************************/
    private $needFlush;

    /***********************************************************************
     * Constructor.
     ***********************************************************************/
    public function __construct(){
        $this->issueChangeSet = [];
    }

    /***********************************************************************
     * Called by Doctrine prePersist event.
     ***********************************************************************/
    public function prePersist(LifecycleEventArgs $eventArgs){}

    /***********************************************************************
     * Called by Doctrine postPersist event.
     ***********************************************************************/
    public function postPersist(LifecycleEventArgs $eventArgs){}

    /***********************************************************************
     * Called by Doctrine preUpdate event.
     ***********************************************************************/
    public function preUpdate(PreUpdateEventArgs $eventArgs){
        $changedEntity = $eventArgs->getEntity();
        if($this->isEntityLogGenerator($changedEntity)){
            $this->saveChangedFields($eventArgs);
        }
    }

    /***********************************************************************
     * Called by Doctrine postUpdate event.
     ***********************************************************************/
    public function postUpdate(LifecycleEventArgs $eventArgs){
        $changedEntity = $eventArgs->getEntity();
        if($this->isEntityLogGenerator($changedEntity)){
            $this->addLog($changedEntity);
        }
    }

    /***********************************************************************
     * Called by Doctrine postFlush event.
     ***********************************************************************/
    public function postFlush(PostFlushEventArgs $eventArgs){
        if($this->needFlush){
            $this->needFlush = false;
            $eventArgs->getEntityManager()->flush();
        }
    }

    /***********************************************************************
     * Gets the methods that will be triggered by the event.
     ***********************************************************************/
    public function getSubscribedEvents(){
        return array(
            Events::prePersist,
            Events::postPersist,
            Events::preUpdate,
            Events::postUpdate,
            Events::postFlush
        );
    }

    /***********************************************************************
     * Check if is a Task entity and it is an issue.
     ***********************************************************************/
    protected function isEntityIssue($entity){
        if($entity instanceof TaskEntity){
            return $entity->isIssue();
        }
        return false;
    }

    /***********************************************************************
     * Check if is an Allocation entity.
     ***********************************************************************/
    protected function isEntityAllocation($entity){
        return $entity instanceof AllocationEntity;
    }

    /***********************************************************************
     * Check if entity's class is available to create issue log.
     ***********************************************************************/
    protected function isEntityLogGenerator($entity){
        return ($this->isEntityAllocation($entity) || $this->isEntityIssue($entity));
    }

    /***********************************************************************
     * Add a new log when issue was changed.
     ***********************************************************************/
    protected function addItemChangeSet($oldValue, $newValue, $tableName, $columnName){
        array_push($this->issueChangeSet, array(
            'OLD_VALUE' => $oldValue,
            'NEW_VALUE' => $newValue,
            'TABLE_NAME' => $tableName,
            'COLUMN_NAME' => $columnName
        ));
    }

    /***********************************************************************
     * Check if field was really changed, because sometimes the old value
     * is null and, the new value, an empty string.
     ***********************************************************************/
    protected function isFieldReallyChanged(PreUpdateEventArgs $eventArgs, $fieldName){
        $oldValue = $eventArgs->getOldValue($fieldName);
        $newValue = $eventArgs->getNewValue($fieldName);
        return !(empty($oldValue) && empty($newValue));
    }

    /**
     * Calls a given value method.
     *
     * @param $value
     * @param $method
     * @throws \Exception
     * @return null
     */
    protected function callObjectMethodIfExists($value, $method){
        if(empty($value)) return null;
        if( ! method_exists($value, $method)){
            throw new \Exception("The given object does not contain $method method.");
        }
        return  $value->$method();
    }

    /***********************************************************************
     * Check changed fields and gets their old and new values.
     ***********************************************************************/
    protected function saveChangedFields(PreUpdateEventArgs $eventArgs){
        $entity = $eventArgs->getEntity();
        $changedTable = $eventArgs->getEntityManager()->getClassMetadata(get_class($entity))->getTableName();

        if($eventArgs->hasChangedField(self::DESCRIPTION_FIELD)){
            if($this->isFieldReallyChanged($eventArgs, self::DESCRIPTION_FIELD)){
                $oldValue = $eventArgs->getOldValue(self::DESCRIPTION_FIELD);
                $newValue = $eventArgs->getNewValue(self::DESCRIPTION_FIELD);
                $this->addItemChangeSet($oldValue, $newValue, $changedTable, "DESCRIPTION");
            }
        }

        if($eventArgs->hasChangedField(self::TASK_STATUS_FIELD)){
            if($this->isFieldReallyChanged($eventArgs, self::TASK_STATUS_FIELD)){
                $oldValue = $this->callObjectMethodIfExists($eventArgs->getOldValue(self::TASK_STATUS_FIELD), "getId");
                $newValue = $this->callObjectMethodIfExists($eventArgs->getNewValue(self::TASK_STATUS_FIELD), "getId");
                $this->addItemChangeSet($oldValue, $newValue, $changedTable, "TASK_STATUS");
            }
        }

        if($eventArgs->hasChangedField(self::REPORTER_FIELD)){
            if($this->isFieldReallyChanged($eventArgs, self::REPORTER_FIELD)){
                $oldValue = $this->callObjectMethodIfExists($eventArgs->getOldValue(self::REPORTER_FIELD), "getId");
                $newValue = $this->callObjectMethodIfExists($eventArgs->getNewValue(self::REPORTER_FIELD), "getId");
                $this->addItemChangeSet($oldValue, $newValue, $changedTable, "REPORTER");
            }
        }
    }

    /***********************************************************************
     * Gets issueChangeSet and sends to Service
     ***********************************************************************/
    protected function addLog($entity){
        /** @var \Service\IssueLog\IssueLog $issueLogService */
        $issueLogService = InstanceManager::getInstance()->getService('\Service\IssueLog');
        foreach($this->issueChangeSet as $changeItem){
            $oldValue   = $changeItem['OLD_VALUE'];
            $newValue   = $changeItem['NEW_VALUE'];
            $tableName  = $changeItem['TABLE_NAME'];
            $columnName = $changeItem['COLUMN_NAME'];
            $issueLogService->logUpdate($entity, $oldValue, $newValue, $tableName, $columnName);
        }
        $this->needFlush = true;
        $this->issueChangeSet = [];
    }


} 